#include "chess_board.h"
#include "stdio.h"
#include <string.h>


static uint8_t Board_Pieces[8][8] = {{0}};
static uint8_t Saved_Binary_Board[8] = {0};
static uint8_t Board_States[8][8] = {{0}};
static Game_State_t Game_State;
static Game_State_t Last_Game_State;
static uint8_t Selected_Piece = SQUARE_EMPTY;
static uint8_t Error_Count = 0u;
static uint8_t Taken_Piece = SQUARE_EMPTY;
static bool Check[2u] = {false, false};
static uint8_t King_Locations[2u][2u] = {{7,4}, {0,4}};
static bool Castling_Allowed[2u][2u] = {{true, true}, {true, true}};
static bool High_Alert = false;


static bool Converting_Pawn = false;
static uint8_t Converting_Pawn_Row_Col[2];
static uint8_t Pawn_Converted_To = QUEEN_WHITE;


/**
 * @brief  Function for clearing all of the lights on the board. Except for error moves.
 * @retval None
 */
static void clear_board_state(void)
{
    for (size_t i = 0; i < 8; i++)
    {
        for (size_t j = 0; j < 8; j++)
        {
            if(Board_States[i][j] != ERROR_MOVE)
            {
                Board_States[i][j] = LIGHT_OFF;
            }
        }
    }

}

/**
 * @brief Function for determining if the piece is on the white team or on the black team.
 * @note   Pieces should be enumerated even if white and odd if black.
 * @param  piece: The piece under question.
 * @retval Return true if on the white team, else false.
 */
static bool white_team(uint8_t piece)
{
    return ((piece % 2u) == 0u);
}

/**
 * @brief  Function for determining if two pieces are on the same team or not.
 * @param  piece_one: Piece one ofcoarse.
 * @param  piece_two: Piece two obviously.
 * @retval True if on opposite teams, else false.
 */
static bool opposite_teams(uint8_t piece_one, uint8_t piece_two)
{
    return (((piece_one % 2u) == 0u) ^ ((piece_two % 2u) == 0u));
}

/**
 * @brief  Check to see if the square is safe from the other team.
 * @param  column: Column of potential move
 * @param  row: Row of the potential move
 * @retval True if the square is safe, else is false
 */
bool square_is_safe(uint8_t row, uint8_t column)
{
    int8_t temp = row + (Game_State.player_turn ? -1 : 1);

    /* first check if pawns can take us */
    if ((column > 0) && ((Game_State.player_turn ? PAWN_BLACK : PAWN_WHITE) == Board_Pieces[temp][column - 1u]))
    {
        //can be eaten by a pawn, not safe.
        return false;
    }
    if ((column < 7) && ((Game_State.player_turn ? PAWN_BLACK : PAWN_WHITE) == Board_Pieces[temp][column + 1u]))
    {
        //can be eaten by a pawn, not safe.
        return false;
    }

    /* Other King */
    int8_t start_r = (row == 0u) ? 0 : -1;
    int8_t stop_r = (row == 7u) ? 0 : 1;
    int8_t start_c = (column == 0u) ? 0 : -1;
    int8_t stop_c = (column == 7u) ? 0 : 1;
    for (int8_t up_down = start_r; up_down <= stop_r; up_down++)
    {
        for (int8_t left_right = start_c; left_right <= stop_c; left_right++)
        {
            int8_t x = row + left_right;
            int8_t y = column + up_down;
            if ((Game_State.player_turn ? KING_BLACK : KING_WHITE) == Board_Pieces[x][y])
            {
                return false;
            }
        }
    }

    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 1;
            left_right_step = 0;
        }
        else if (direction == 1u)
        {
            up_down_step = 0;
            left_right_step = 1;
        }
        else if (direction == 2u)
        {
            up_down_step = -1;
            left_right_step = 0;
        }
        else
        {
            up_down_step = 0;
            left_right_step = -1;
        }
        /* Rooks and queens */
        bool loop = true;
        int8_t x = row;
        int8_t y = column;
        while (loop)
        {
            x += left_right_step;
            y += up_down_step;
            if ((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
            {
                loop = false;
            }
            else if (Board_Pieces[x][y] == SQUARE_EMPTY)
            {
                /* Do nothing */
            }
            else if (((Game_State.player_turn ? ROOK_BLACK : ROOK_WHITE) == Board_Pieces[x][y]) || ((Game_State.player_turn ? QUEEN_BLACK : QUEEN_WHITE) == Board_Pieces[x][y]))
            {
                return false;
            }
            else
            {
                loop = false;
            }
        }
    }
    /* Bishops and queens */
    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 1;
            left_right_step = 1;
        }
        else if (direction == 1u)
        {
            up_down_step = -1;
            left_right_step = 1;
        }
        else if (direction == 2u)
        {
            up_down_step = -1;
            left_right_step = -1;
        }
        else
        {
            up_down_step = 1;
            left_right_step = -1;
        }
        bool loop = true;
        int8_t x = row;
        int8_t y = column;
        while (loop)
        {
            uint8_t bish = (Game_State.player_turn ? BISHOP_BLACK : BISHOP_WHITE);
            uint8_t queen = (Game_State.player_turn ? QUEEN_BLACK : QUEEN_WHITE);
            x += left_right_step;
            y += up_down_step;
            if ((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
            {
                loop = false;
            }
            else if (Board_Pieces[x][y] == SQUARE_EMPTY)
            {
                /* do nothing */
            }
            else if ((bish == Board_Pieces[x][y]) || (queen == Board_Pieces[x][y]))
            {
                return false;
            }
            else
            {
                loop = false;
            }
        }
    }
    /* Knights */
    for (uint8_t direction = 0u; direction < 4u; direction++)
    {
        int8_t up_down_step;
        int8_t left_right_step;
        if (direction == 0u)
        {
            up_down_step = 2;
            left_right_step = 0;
        }
        else if (direction == 1u)
        {
            up_down_step = 0;
            left_right_step = 2;
        }
        else if (direction == 2u)
        {
            up_down_step = -2;
            left_right_step = 0;
        }
        else
        {
            up_down_step = 0;
            left_right_step = -2;
        }

        for (uint8_t i = 0u; i < 2u; i++)
        {
            if ((direction % 2u) == 0u)
            {
                left_right_step = (i == 0u) ? -1 : 1;
            }
            else
            {
                up_down_step = (i == 0u) ? -1 : 1;
            }

            int8_t x = (int8_t)row + left_right_step;
            int8_t y = (int8_t)column + up_down_step;
            if ((x >= 0) && (y >= 0) && (x < 8) && (y < 8))
            {
                if ((Game_State.player_turn ? KNIGHT_BLACK : KNIGHT_WHITE) == Board_Pieces[x][y])
                {
                    return false;
                }
            }
        }
    }
    return true;
}

bool Check_If_Move_Caused_Check(uint8_t piece, uint8_t row, uint8_t column)
{
    bool ret_val;
    uint8_t store_current_piece = Board_Pieces[row][column];
    Board_Pieces[row][column] = piece;
    //If its the white's turn we want to see if the white king is still safe.
    uint8_t white_black_idx = Game_State.player_turn ? 0u : 1u;
    ret_val = !square_is_safe(King_Locations[white_black_idx][0u], King_Locations[white_black_idx][1u]);
    Board_Pieces[row][column] = store_current_piece;
    return ret_val;
}

void Check_If_Could_Cause_Check(uint8_t row, uint8_t column)
{
    uint8_t temp_storage = Board_Pieces[row][column];
    Board_Pieces[row][column] = SQUARE_EMPTY;
    //If its the white's turn we want to see if the white king is still safe.
    uint8_t white_black_idx = Game_State.player_turn ? 0u : 1u;
    High_Alert = !square_is_safe(King_Locations[white_black_idx][0u], King_Locations[white_black_idx][1u]);
    Board_Pieces[row][column] = temp_storage;
}

static bool Set_Light(uint8_t piece, uint8_t row, uint8_t column, uint8_t state)
{
    bool ret_val = false;
    if (!Check_If_Move_Caused_Check(piece, row, column))
    {
        Board_States[row][column] = state;
        ret_val = true;
    }
    return ret_val;
}

/**
 * @brief  Function for marking potential moves for pawns.
 * @param  row: row to move to
 * @param  column: column to move to
 * @retval None
 */
static bool pawn_move(uint8_t piece, uint8_t row, uint8_t column)
{
    bool ret_val = false;
    if (Board_Pieces[row][column] == SQUARE_EMPTY)
    {
        ret_val = Set_Light(piece, row, column, POTENTIAL_MOVE);
    }
    return ret_val;
}

/**
 * @brief  Function for "casting" a ray in any direction, vertical, horizontal, or diagonal. If the ray hits someone from the other team
 *         or the end of the board the array will be terminated.
 * @param  direction_r: Row direction
 * @param  direction_c: Column direction
 * @param  row: current row location
 * @param  column: current column location
 * @param  piece: the piece that is casting the ray.
 * @retval None
 */
static bool cast_a_ray(uint8_t piece, int8_t direction_r, int8_t direction_c, uint8_t column, uint8_t row)
{
    bool ret_val = false;
    bool loop = true;
    int8_t x = row;
    int8_t y = column;
    while (loop)
    {
        x += direction_c;
        y += direction_r;
        if((x < 0) || (y < 0) || (x >= 8) || (y >= 8))
        {
            loop = false;
        }
        else if (Board_Pieces[x][y] == SQUARE_EMPTY)
        {
            ret_val = Set_Light(piece, x, y, POTENTIAL_MOVE) || ret_val;
        }
        else if (opposite_teams(piece, Board_Pieces[x][y]))
        {
            ret_val = Set_Light(piece, x, y, POTENTIAL_TAKE) || ret_val;
            /* once we take a piece we can no longer take anymore */
            loop = false;
        }
        else
        {
            loop = false;
        }
    }
    return ret_val;
}

/**
 * @brief  Marking a piece for taking.
 * @param  row: Row ofcoarse
 * @param  column: Column obviously
 * @retval None
 */
static bool pawn_take(uint8_t piece, uint8_t row, uint8_t column)
{
    bool ret_val = false;
    if ((Board_Pieces[row][column] < SQUARE_EMPTY) && opposite_teams(piece, Board_Pieces[row][column]))
    {
        ret_val = Set_Light(piece, row, column, POTENTIAL_TAKE);
    }
    return ret_val;
}

/**
 * @brief  Function for marking the potential moves.
 * @param  piece: Piece that we are marking the potential moves for.
 * @param  row: Current row location of the piece.
 * @param  column: Current column location of the piece.
 * @retval None
 */
static bool Mark_Potential_Moves(uint8_t piece, uint8_t column, uint8_t row)
{
    bool ret_val = false;
    switch (piece)
    {
        case PAWN_WHITE :
        case PAWN_BLACK:
        {
            int8_t direction = white_team(piece) ? -1 : 1;
            uint8_t temp_row = row + direction;
            if (row == (white_team(piece) ? 6u : 1u))
            {
                if(Board_Pieces[temp_row][column] == SQUARE_EMPTY)
                {
                    ret_val = pawn_move(piece, row + (direction * 2u), column) || ret_val;
                }
            }

            ret_val = pawn_move(piece, temp_row, column) || ret_val;
            if(column != 0)
            {
                ret_val = pawn_take(piece, temp_row, column - 1u) || ret_val;
            }
            if(column != 7u)
            {
                ret_val = pawn_take(piece, temp_row, column + 1u) || ret_val;
            }

            break;
        }
        case ROOK_WHITE:
        case ROOK_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if(direction == 0u)
                {
                    up_down_step = 1;
                    left_right_step = 0;
                }
                else if(direction == 1u)
                {
                    up_down_step = 0;
                    left_right_step = 1;
                }
                else if (direction == 2u)
                {
                    up_down_step = -1;
                    left_right_step = 0;
                }
                else
                {
                    up_down_step = 0;
                    left_right_step = -1;
                }
                ret_val = cast_a_ray(piece, up_down_step, left_right_step, column, row) || ret_val;
            }
            break;

        }
        case KNIGHT_WHITE:
        case KNIGHT_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if (direction == 0u)
                {
                    up_down_step = 2;
                    left_right_step = 0;
                }
                else if (direction == 1u)
                {
                    up_down_step = 0;
                    left_right_step = 2;
                }
                else if (direction == 2u)
                {
                    up_down_step = -2;
                    left_right_step = 0;
                }
                else
                {
                    up_down_step = 0;
                    left_right_step = -2;
                }

                for (uint8_t i = 0u; i < 2u; i++)
                {
                    if((direction % 2u) == 0u)
                    {
                        left_right_step = (i == 0u) ? -1 : 1;
                    }
                    else
                    {
                        up_down_step = (i == 0u) ? -1 : 1;
                    }

                    int8_t x = (int8_t)row + left_right_step;
                    int8_t y = (int8_t)column + up_down_step;
                    if ((x >= 0) && (y >= 0) && (x < 8) && (y < 8))
                    {
                        if (Board_Pieces[x][y] == SQUARE_EMPTY)
                        {
                            ret_val = Set_Light(piece, x, y, POTENTIAL_MOVE) || ret_val;
                        }
                        else if (opposite_teams(piece, Board_Pieces[x][y]))
                        {
                            ret_val = Set_Light(piece, x, y, POTENTIAL_TAKE) || ret_val;
                        }
                    }
                }
            }
            break;
        }
        case BISHOP_WHITE:
        case BISHOP_BLACK:
        {
            for (uint8_t direction = 0u; direction < 4u; direction++)
            {
                int8_t up_down_step;
                int8_t left_right_step;
                if (direction == 0u)
                {
                    up_down_step = 1;
                    left_right_step = 1;
                }
                else if (direction == 1u)
                {
                    up_down_step = -1;
                    left_right_step = 1;
                }
                else if (direction == 2u)
                {
                    up_down_step = -1;
                    left_right_step = -1;
                }
                else
                {
                    up_down_step = 1;
                    left_right_step = -1;
                }
                ret_val = cast_a_ray(piece, up_down_step, left_right_step, column, row) || ret_val;
            }
            break;
        }
        case QUEEN_WHITE:
        case QUEEN_BLACK:
        {
            //Mark bishop moves
            ret_val = Mark_Potential_Moves((piece - 2u), column, row) || ret_val;
            //Mark rook moves
            ret_val = Mark_Potential_Moves((piece - 6u), column, row) || ret_val;
            break;
        }
        case KING_WHITE:
        case KING_BLACK:
        {
            int8_t start_r = (row == 0u) ? 0 : -1;
            int8_t stop_r = (row == 7u) ? 0 : 1;
            int8_t start_c = (column == 0u) ? 0 : -1;
            int8_t stop_c = (column == 7u) ? 0 : 1;
            for (int8_t left_right = start_r; left_right <= stop_r; left_right++)
            {
                for (int8_t up_down = start_c; up_down <= stop_c; up_down++)
                {
                    int8_t x = row + left_right;
                    int8_t y = column + up_down;
                    if (square_is_safe(x, y))
                    {
                        if (Board_Pieces[x][y] == SQUARE_EMPTY)
                        {
                            Board_States[x][y] = POTENTIAL_MOVE;
                            ret_val = true;
                        }
                        else if (opposite_teams(piece, Board_Pieces[x][y]))
                        {
                            Board_States[x][y] = POTENTIAL_TAKE;
                            ret_val = true;
                        }
                    }
                }
            }
            uint8_t white_black_idx = white_team(piece) ? 0u : 1u;
            uint8_t kings_row = white_team(piece) ? 7u : 0u;

            //Can only castle if not currently in check
            if (square_is_safe(row, column))
            {
                // Queen side castle
                if(Castling_Allowed[white_black_idx][0u] && (Board_Pieces[kings_row][1u] == SQUARE_EMPTY)
                && (Board_Pieces[kings_row][2u] == SQUARE_EMPTY) && (Board_Pieces[kings_row][3u]) == SQUARE_EMPTY)
                {
                    //First Check to see if the king will pass through check
                    if(square_is_safe(kings_row, 3u) && square_is_safe(kings_row, 2u))
                    {
                        // Yay we can castle queen side!
                        Board_States[kings_row][2u] = POTENTIAL_CASTLE;
                        ret_val = true;
                    }
                }

                // King side castle
                if (Castling_Allowed[white_black_idx][1u] && (Board_Pieces[kings_row][5u] == SQUARE_EMPTY) && (Board_Pieces[kings_row][6u] == SQUARE_EMPTY))
                {
                    //First Check to see if the king will pass through check
                    if(square_is_safe(kings_row, 5u) && square_is_safe(kings_row, 6u))
                    {
                        // Yay we can castle king side!
                        Board_States[kings_row][6u] = POTENTIAL_CASTLE;
                        ret_val = true;
                    }
                }
            }

            break;
        }

        default:
            break;
    }
    return ret_val;
}

bool Check_If_Player_Can_Move(bool white)
{
    for (uint8_t row = 0; row < 8u; row++)
    {
        for (uint8_t column = 0; column < 8u; column++)
        {
            if((white_team(Board_Pieces[row][column]) == white))
            {
                // SDL_Log("move: Row:%d, Col:%d", row, column);

                if(Mark_Potential_Moves(Board_Pieces[row][column], column, row))
                {
                    // SDL_Log("Player Can still move: Row:%d, Col:%d", row, column);
                    clear_board_state();
                    return true;
                }
            }
        }

    }
    clear_board_state();
    // SDL_Log("Player cant move");
    return false;
}

/**
 * @brief  Function for switching the players turn. Incharge of handling the state machine reset.
 */
static void Switch_Turns(void)
{
    Game_State.turn_state = BEGINNING;
    Game_State.player_turn = !Game_State.player_turn;
    // Square is safe assumes the other team is trying to attack the square so for example at the end of
    // White's turn we want to see if the black king is now in check, so we will switch teams and then
    // Check if the current kings locations is safe. If it is safe then check is false, if it isnt safe then check is true.
    uint8_t white_black_idx = Game_State.player_turn ? 0u : 1u;
    Check[white_black_idx] = !square_is_safe(King_Locations[white_black_idx][0u], King_Locations[white_black_idx][1u]);

    //Last thing we need to check before switching turns is to check if the game is over.
    bool player_can_play = Check_If_Player_Can_Move(Game_State.player_turn);
    if(!player_can_play)
    {
        Game_State.game_over = true;
        Game_State.player_turn = !Game_State.player_turn;
        Game_State.error_detected = !Check[white_black_idx];
    }

}

/**
 * @brief  Function for checking the selected piece to see if we are moving the king.
 *         If we are then we also want to update the new location of the corresponding king.
 * @param  row: Current row location of the piece.
 * @param  column: Current column location of the piece.
 * @retval None
 */
static void Check_If_Moving_King(uint8_t row, uint8_t column)
{
    uint8_t white_black_idx = Game_State.player_turn ? 0u : 1u;
    if((Selected_Piece == KING_WHITE) || (Selected_Piece == KING_BLACK))
    {
        King_Locations[white_black_idx][0u] = row;
        King_Locations[white_black_idx][1u] = column;
        Castling_Allowed[white_black_idx][0u] = false;
        Castling_Allowed[white_black_idx][1u] = false;
    }
    // Disable the castling of the corresponding side if the rook is being moved.
    else if (((Selected_Piece == ROOK_WHITE) && (row == 7u))
        || ((Selected_Piece == ROOK_BLACK) && (row == 0u)))
    {
        if (column == 0u)
        {
            Castling_Allowed[white_black_idx][0u] = false;
        }
        else if (column == 7u)
        {
            Castling_Allowed[white_black_idx][1u] = false;
        }
    }
}

static void Check_If_Converting_Pawn(uint8_t row, uint8_t column)
{
    uint8_t white_black_idx = Game_State.player_turn ? 0u : 1u;
    Converting_Pawn = false;

    if((Selected_Piece == PAWN_WHITE) || (Selected_Piece == PAWN_BLACK))
    {
        if((row == 0u) || (row == 7u))
        {
            Selected_Piece = Game_State.player_turn ? QUEEN_WHITE : QUEEN_BLACK;
            Pawn_Converted_To = Selected_Piece;
            Converting_Pawn = true;
            Converting_Pawn_Row_Col[0] = row;
            Converting_Pawn_Row_Col[1] = column;
        }
    }
}

static bool Converting_Pawn_If_Applicable(uint8_t row, uint8_t column)
{
    bool ret_val = false;
    if(Converting_Pawn)
    {
        if((row == Converting_Pawn_Row_Col[0]) &&
            (Converting_Pawn_Row_Col[1] == column))
        {
            //Putting the piece down on the board
            if(Board_Pieces[row][column] == SQUARE_EMPTY)
            {
                Board_Pieces[row][column] = Pawn_Converted_To;
                Board_States[row][column] = LIGHT_OFF;
            }
            //Picking the piece back up to toggle through the options
            else
            {
                Board_Pieces[row][column] = SQUARE_EMPTY;
                Board_States[row][column] = CONVERTING_PAWN;
                Pawn_Converted_To = Pawn_Converted_To - 2;
                if (Pawn_Converted_To < ROOK_WHITE)
                {
                    Pawn_Converted_To += 8u;
                }

            }
            ret_val = true;
        }

    }

    return ret_val;
}


/**
 * @brief  Function for toggeling a square's state.
 * @param  j: row location that was toggled.
 * @param  i: column location that was toggled.
 * @retval None
 */
static void Board_Square_Was_Toggled(uint8_t j, uint8_t i)
{
    if (Game_State.error_detected)
    {
        if (Board_States[j][i] == PIECE_ORIGIN)
        {
            Board_Pieces[j][i] = Selected_Piece;
            Selected_Piece = SQUARE_EMPTY;
            clear_board_state();
            Game_State.turn_state = BEGINNING;
        }
        else if (Board_States[j][i] == PIECE_NEEDS_TO_BE_HERE)
        {
            if (j < 8u)
            {
                Board_Pieces[j][i] = Selected_Piece;
                Selected_Piece = SQUARE_EMPTY;
                Board_States[j][i] = LIGHT_OFF;
            }
            else
            {
                Board_Pieces[j][i] = Taken_Piece;
                uint8_t board_column = (j / 2u) * 2u;
                if(Board_States[board_column][i] == PIECE_NEEDS_TO_BE_HERE)
                {
                    Board_States[board_column][i] = LIGHT_OFF;
                }
                if(Board_States[board_column + 1u][i] == PIECE_NEEDS_TO_BE_HERE)
                {
                    Board_States[board_column + 1u][i] = LIGHT_OFF;
                }
                Taken_Piece = SQUARE_EMPTY;
            }

            if ((Selected_Piece == SQUARE_EMPTY) && (Taken_Piece == SQUARE_EMPTY))
            {
                Game_State.turn_state = BEGINNING;
                Game_State.player_turn = !Game_State.player_turn;
            }
        }
        else if (Board_States[j][i] == ERROR_MOVE)
        {
            Error_Count--;
            Board_States[j][i] = LIGHT_OFF;
            /* All Errors have been rectified so we can go back to where we were.*/
            if(Error_Count == 0u)
            {
                Game_State.error_detected = false;
            }
        }
        else
        {
            Error_Count++;
            Board_States[j][i] = ERROR_MOVE;
        }

    }
    else
    {
        switch (Game_State.turn_state)
        {
            /* We are waiting till the player who's turn it is picks up a piece that is on their team */
            case BEGINNING:
            {
                if ((j < 8u) && (Board_Pieces[j][i] != SQUARE_EMPTY) && (white_team(Board_Pieces[j][i]) == Game_State.player_turn))
                {
                    if((Board_Pieces[j][i] != KING_BLACK) && (Board_Pieces[j][i] != KING_WHITE))
                    {
                        Check_If_Could_Cause_Check(j, i);
                    }
                    Selected_Piece = Board_Pieces[j][i];
                    Board_Pieces[j][i] = SQUARE_EMPTY;
                    (void)Mark_Potential_Moves(Selected_Piece, i, j);
                    Board_States[j][i] = PIECE_ORIGIN;
                    Game_State.turn_state = IN_PROGRESS;
                }
                else
                {
                    if(!Converting_Pawn_If_Applicable(j, i))
                    {
                        Game_State.error_detected = true;
                        Board_States[j][i] = ERROR_MOVE;
                        Error_Count++;
                    }
                }

                break;
            }
            /* Person is in the middle of taking a turn for example they might already have a piece in the hand*/
            case IN_PROGRESS:
            {
                if (Board_States[j][i] == POTENTIAL_MOVE)
                {
                    Check_If_Moving_King(j, i);
                    Check_If_Converting_Pawn(j, i);
                    Board_Pieces[j][i] = Selected_Piece;
                    Selected_Piece = SQUARE_EMPTY;
                    clear_board_state();
                    Switch_Turns();
                }
                else if (Board_States[j][i] == POTENTIAL_TAKE)
                {
                    // Taken_Piece = Board_State[j][i];
                    Board_Pieces[j][i] = SQUARE_EMPTY;
                    Game_State.turn_state = FINALIZING;
                    clear_board_state();
                    Board_States[j][i] = PIECE_NEEDS_TO_BE_HERE;
                }
                else if (Board_States[j][i] == PIECE_ORIGIN)
                {
                    Board_Pieces[j][i] = Selected_Piece;
                    Selected_Piece = SQUARE_EMPTY;
                    clear_board_state();
                    Game_State.turn_state = BEGINNING;
                }
                else if (Board_States[j][i] == POTENTIAL_CASTLE)
                {
                    Check_If_Moving_King(j, i);
                    Board_Pieces[j][i] = Selected_Piece;
                    Selected_Piece = SQUARE_EMPTY;
                    clear_board_state();
                    if(i == 2u)
                    {
                        Board_States[j][3u] = PIECE_NEEDS_TO_BE_HERE;
                        Board_States[j][0u] = PIECE_NEEDS_TO_BE_REMOVED;
                    }
                    else if(i == 6u)
                    {
                        Board_States[j][5u] = PIECE_NEEDS_TO_BE_HERE;
                        Board_States[j][7u] = PIECE_NEEDS_TO_BE_REMOVED;
                    }
                    else
                    {
                        /* Do nothing. */
                    }

                }
                else if (Board_States[j][i] == PIECE_NEEDS_TO_BE_REMOVED)
                {
                    Selected_Piece = Board_Pieces[j][i];
                    Board_Pieces[j][i] = SQUARE_EMPTY;
                    Board_States[j][i] = LIGHT_OFF;
                    Game_State.turn_state = FINALIZING;
                }
                else
                {
                    Game_State.error_detected = true;
                    Board_States[j][i] = ERROR_MOVE;
                    Error_Count++;
                }
                break;
            }
            /* Player still needs to do something to complete their turn, like complete castle, en pessant, or converting a pawn*/
            case FINALIZING:
            {
                if (Board_States[j][i] == PIECE_NEEDS_TO_BE_HERE)
                {
                    Check_If_Moving_King(j, i);
                    Check_If_Converting_Pawn(j, i);
                    Board_Pieces[j][i] = Selected_Piece;
                    Selected_Piece = SQUARE_EMPTY;
                    Board_States[j][i] =  LIGHT_OFF;
                }
                else
                {
                    if(!Converting_Pawn_If_Applicable(j, i))
                    {
                        Game_State.error_detected = true;
                        Board_States[j][i] = ERROR_MOVE;
                        Error_Count++;
                    }
                }

                if ((Selected_Piece == SQUARE_EMPTY) && (Taken_Piece == SQUARE_EMPTY))
                {
                    Switch_Turns();
                }

                break;
            }
            default:
            {
                break;
            }
        }
    }
}

void Board_get_lights_and_state(uint8_t board_state[8][8], uint8_t board_pieces[8][8])
{
    memcpy(&board_state[0][0], &Board_States[0][0], sizeof(Board_States));
    memcpy(&board_pieces[0][0], &Board_Pieces[0][0], sizeof(Board_Pieces));
}


/**
 * @brief  The Board changed so now we have to see what is different and act accordingly.
 * @note   Yes i know the design of this seems really bad but it's important to remember this is supposed to simulate the chess board I'm creating.
 *         so I'm designing it this way because of the hardware that I'm using.
 * @retval None
 */
void Board_Changed(uint8_t current_binary_board[8])
{
    for (uint8_t j = 0u; j < 8u; j++)
    {
        uint8_t difference = (current_binary_board[j] ^ Saved_Binary_Board[j]);
        if (difference != 0u)
        {
            for (uint8_t i = 0u; i < 8u; i++)
            {
                if((difference & (1u << i)) != 0u)
                {
                    Board_Square_Was_Toggled(j, i);
                }
            }
        }
        Saved_Binary_Board[j] = current_binary_board[j];
    }
}


/**
 * @brief  Function for initializing the board
 * @note
 * @retval None
 */
void chess_board_init(void)
{
    for (uint8_t i = 0u; i < 8u; i++)
    {
        for (uint8_t j = 0u; j < 8u; j++)
        {
            Board_Pieces[i][j] = SQUARE_EMPTY;
        }
    }
    Saved_Binary_Board[0] = 0xFF;
    Saved_Binary_Board[1] = 0xFF;
    Saved_Binary_Board[6] = 0xFF;
    Saved_Binary_Board[7] = 0xFF;

    //Place black pieces
    Board_Pieces[0][0] = ROOK_BLACK;
    Board_Pieces[0][7] = ROOK_BLACK;
    Board_Pieces[0][1] = KNIGHT_BLACK;
    Board_Pieces[0][6] = KNIGHT_BLACK;
    Board_Pieces[0][2] = BISHOP_BLACK;
    Board_Pieces[0][5] = BISHOP_BLACK;
    Board_Pieces[0][3] = QUEEN_BLACK;
    Board_Pieces[0][4] = KING_BLACK;
    Board_Pieces[7][0] = ROOK_WHITE;
    Board_Pieces[7][7] = ROOK_WHITE;
    Board_Pieces[7][1] = KNIGHT_WHITE;
    Board_Pieces[7][6] = KNIGHT_WHITE;
    Board_Pieces[7][2] = BISHOP_WHITE;
    Board_Pieces[7][5] = BISHOP_WHITE;
    Board_Pieces[7][3] = QUEEN_WHITE;
    Board_Pieces[7][4] = KING_WHITE;

    for (uint8_t i = 0; i < 8; i++)
    {
        Board_Pieces[1][i] = PAWN_BLACK;
        Board_Pieces[6][i] = PAWN_WHITE;
    }
}

Game_State_t Board_get_game_state(void)
{
    return Game_State;
}
